"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.PicoEngineCore = void 0;
const pico_framework_1 = require("pico-framework");
const CorePico_1 = require("./CorePico");
const formatEventOutput_1 = require("./formatEventOutput");
const makeKrlCtx_1 = require("./makeKrlCtx");
const ctx_1 = require("./modules/ctx");
const event_1 = require("./modules/event");
const http_1 = require("./modules/http");
const math_1 = require("./modules/math");
const random_1 = require("./modules/random");
const schedule_1 = require("./modules/schedule");
const stdlib_1 = require("./modules/stdlib");
const time_1 = require("./modules/time");
const ursa_1 = require("./modules/ursa");
const dido_1 = require("./modules/dido");
const meta_1 = require("./modules/meta");
const RulesetRegistry_1 = require("./RulesetRegistry");
const version = require("../package.json").version;
class PicoEngineCore {
    constructor(conf) {
        this.version = version;
        this.modules = {};
        this.picos = {};
        this.startupModules = [];
        const log = (this.log = conf.log);
        this.getPicoLogs = conf.getPicoLogs;
        this.rsRegistry = new RulesetRegistry_1.RulesetRegistry(conf.rsRegLoader, (crs) => this.onRulesetLoaded(crs));
        this.picoFramework = new pico_framework_1.PicoFramework({
            leveldown: conf.leveldown,
            genID: conf.genID,
            useEventInputTime: conf.useEventInputTime,
            environment: (rsCtx) => (0, makeKrlCtx_1.makeKrlCtx)(this, rsCtx),
            rulesetLoader: this.rsRegistry.loader,
            onFrameworkEvent(ev) {
                switch (ev.type) {
                    case "startup":
                        break;
                    case "startupDone":
                        log.debug("pico-framework started");
                        break;
                    case "startupRulesetInitError":
                        log.error("onStartupRulesetInitError", {
                            picoId: ev.picoId,
                            rid: ev.rid,
                            rulesetConfig: ev.config,
                            error: ev.error,
                        });
                        break;
                    case "txnQueued":
                        log.debug(ev.type, {
                            picoId: ev.picoId,
                            txnId: ev.txn.id,
                            txn: ev.txn,
                        });
                        break;
                    case "txnStart":
                    case "txnDone":
                        log.debug(ev.type, { picoId: ev.picoId, txnId: ev.txn.id });
                        break;
                    case "txnError":
                        log.debug(ev.type, {
                            picoId: ev.picoId,
                            txnId: ev.txn.id,
                            error: ev.error + "",
                        });
                        break;
                    case "eventScheduleAdded":
                        log.debug("event added to schedule", {
                            picoId: ev.picoId,
                            txnId: ev.txn.id,
                            rid: ev.rid,
                            event: ev.event,
                        });
                        break;
                    case "eventScheduleCleared":
                        log.debug("schedule cleared", {
                            picoId: ev.picoId,
                            txnId: ev.txn.id,
                        });
                        break;
                }
            },
        });
        this.modules["ctx"] = (0, ctx_1.default)(this);
        this.modules["event"] = event_1.default;
        this.modules["http"] = (0, http_1.default)(this);
        this.modules["math"] = math_1.default;
        this.modules["random"] = random_1.default;
        const scheduler = (0, schedule_1.initScheduleModule)(this.picoFramework);
        this.modules["schedule"] = scheduler.module;
        this.startupModules.push(scheduler.start);
        this.modules["stdlib"] = stdlib_1.default;
        this.modules["time"] = time_1.default;
        this.modules["ursa"] = ursa_1.default;
        this.modules["dido"] = dido_1.default;
        this.modules["meta"] = (0, meta_1.default)(this);
        if (conf.modules) {
            for (const domain of Object.keys(conf.modules)) {
                this.modules[domain] = conf.modules[domain];
            }
        }
        this.base_url = conf.base_url;
    }
    async start() {
        for (const startModule of this.startupModules) {
            await startModule();
        }
        await this.picoFramework.start();
    }
    onRulesetLoaded(crs) {
        this.picoFramework.reInitRuleset(crs.ruleset);
        for (const picoId of Object.keys(this.picos)) {
            this.picos[picoId].onRulesetLoaded(crs);
        }
    }
    addPico(picoId) {
        if (!this.picos[picoId]) {
            this.picos[picoId] = new CorePico_1.CorePico(this);
        }
        return this.picos[picoId];
    }
    getPico(picoId) {
        return this.picos[picoId] || null;
    }
    event(event) {
        return this.picoFramework.event(event);
    }
    eventWait(event) {
        return this.picoFramework.eventWait(event).then(formatEventOutput_1.formatEventOutput);
    }
    eventQuery(event, query) {
        return this.picoFramework.eventQuery(event, query);
    }
    query(query) {
        return this.picoFramework.query(query);
    }
}
exports.PicoEngineCore = PicoEngineCore;
//# sourceMappingURL=PicoEngineCore.js.map