"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.RulesetRegistry = void 0;
const crypto = require("crypto");
const pMemoize = require("p-memoize");
class RulesetRegistry {
    constructor(regLoader, onRulesetLoaded) {
        this.regLoader = regLoader;
        this.onRulesetLoaded = onRulesetLoaded;
        this.rulesetCache = {};
        this.loader = async (picoId, rid, config) => {
            const url = config && config["url"];
            if (typeof url !== "string") {
                throw new Error(`Unable to get url for pico ruleset ${picoId}-${rid}`);
            }
            const rs = await this.load(url);
            return rs.ruleset;
        };
        this.load = pMemoize((url) => this.loadBase(url), {
            maxAge: 100,
        });
        this.flush = pMemoize((url) => this.flushBase(url), {
            maxAge: 100,
        });
    }
    getCached(url) {
        return this.rulesetCache[url] || null;
    }
    setCachedRuleset(rs) {
        this.rulesetCache[rs.url] = rs;
        if (this.onRulesetLoaded) {
            this.onRulesetLoaded(rs);
        }
    }
    async loadBase(url) {
        if (this.rulesetCache[url]) {
            return this.rulesetCache[url];
        }
        const data = await this.regLoader.attemptLoad(url);
        if (data) {
            this.setCachedRuleset(data);
            return this.rulesetCache[url];
        }
        return this.flush(url);
    }
    async flushBase(url) {
        var _a;
        const krl = await this.regLoader.fetchKrl(url);
        const shasum = crypto.createHash("sha256");
        shasum.update(krl);
        const hash = shasum.digest("hex");
        if (((_a = this.rulesetCache[url]) === null || _a === void 0 ? void 0 : _a.hash) === hash) {
            return this.rulesetCache[url];
        }
        const { ruleset, compiler } = await this.regLoader.compileAndLoad(url, krl, hash);
        const toSave = {
            url,
            krl,
            hash,
            flushed: new Date(),
            rid: ruleset.rid,
            compiler,
            ruleset,
        };
        await this.regLoader.save(toSave);
        this.setCachedRuleset(toSave);
        return toSave;
    }
}
exports.RulesetRegistry = RulesetRegistry;
//# sourceMappingURL=RulesetRegistry.js.map