"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const cross_fetch_1 = require("cross-fetch");
const krl_stdlib_1 = require("krl-stdlib");
const normalizeUrl = require("normalize-url");
const pico_framework_1 = require("pico-framework");
function initCtxModule(core) {
    function rulesetKrlMetaProperty(key) {
        return krl_stdlib_1.krl.Property(function () {
            const url = this.rsCtx.ruleset.config["url"];
            const cached = core.rsRegistry.getCached(url);
            const krlMeta = (cached === null || cached === void 0 ? void 0 : cached.ruleset).meta || {};
            return krlMeta[key] || null;
        });
    }
    const module = {
        picoId: krl_stdlib_1.krl.Property(function rid() {
            return this.rsCtx.pico().id;
        }),
        rid: krl_stdlib_1.krl.Property(function rid() {
            return this.rsCtx.ruleset.rid;
        }),
        rid_url: krl_stdlib_1.krl.Property(function rid_url() {
            return this.rsCtx.ruleset.config["url"];
        }),
        rid_config: krl_stdlib_1.krl.Property(function rid_config() {
            return this.rsCtx.ruleset.config["config"];
        }),
        rid_version: rulesetKrlMetaProperty("version"),
        rid_name: rulesetKrlMetaProperty("name"),
        rid_description: rulesetKrlMetaProperty("description"),
        rid_author: rulesetKrlMetaProperty("author"),
        parent: krl_stdlib_1.krl.Property(function parent() {
            return this.rsCtx.pico().parent;
        }),
        children: krl_stdlib_1.krl.Property(function children() {
            return this.rsCtx.pico().children;
        }),
        newPico: krl_stdlib_1.krl.Action(["rulesets"], async function newPico(rulesets) {
            if (!Array.isArray(rulesets)) {
                throw new TypeError("ctx:newPico expects an array of {url, config}");
            }
            const toInstall = [];
            for (const rs of rulesets) {
                const result = await core.rsRegistry.load(rs.url);
                toInstall.push({
                    rs: result.ruleset,
                    config: {
                        url: rs.url,
                        config: rs.config || {},
                    },
                });
            }
            const newEci = await this.rsCtx.newPico({
                rulesets: toInstall,
            });
            return newEci;
        }),
        delPico: krl_stdlib_1.krl.Action(["eci"], async function delPico(eci) {
            await this.rsCtx.delPico(eci);
        }),
        channels: krl_stdlib_1.krl.Property(function channels() {
            return this.rsCtx.pico().channels;
        }),
        newChannel: krl_stdlib_1.krl.Action(["tags", "eventPolicy", "queryPolicy"], async function newChannel(tags, eventPolicy, queryPolicy) {
            const conf = { tags, eventPolicy, queryPolicy };
            const chann = await this.rsCtx.newChannel(conf);
            return chann;
        }),
        putChannel: krl_stdlib_1.krl.Action(["eci", "tags", "eventPolicy", "queryPolicy"], async function putChannel(eci, tags, eventPolicy, queryPolicy) {
            const conf = { tags, eventPolicy, queryPolicy };
            await this.rsCtx.putChannel(eci, conf);
        }),
        upsertChannel: krl_stdlib_1.krl.Action(["tags", "eventPolicy", "queryPolicy"], async function upsertChannel(tags, eventPolicy, queryPolicy) {
            tags = (0, pico_framework_1.cleanChannelTags)(tags);
            tags.sort();
            const search = tags.join(",");
            const channel = this.rsCtx.pico().channels.find((c) => {
                return search === c.tags.slice(0).sort().join(",");
            });
            const conf = { tags, eventPolicy, queryPolicy };
            if (channel) {
                await this.rsCtx.putChannel(channel.id, conf);
            }
            else {
                await this.rsCtx.newChannel(conf);
            }
        }),
        delChannel: krl_stdlib_1.krl.Action(["eci"], async function delChannel(eci) {
            await this.rsCtx.delChannel(eci);
        }),
        rulesets: krl_stdlib_1.krl.Property(function rulesets() {
            const pico = this.rsCtx.pico();
            const results = [];
            for (const rs of pico.rulesets) {
                const cached = core.rsRegistry.getCached(rs.config.url);
                results.push({
                    rid: rs.rid,
                    url: rs.config.url,
                    config: rs.config.config,
                    meta: cached
                        ? {
                            krl: cached.krl,
                            krlMeta: (cached === null || cached === void 0 ? void 0 : cached.ruleset).meta,
                            hash: cached.hash,
                            flushed: cached.flushed,
                            compiler: cached.compiler,
                        }
                        : null,
                });
            }
            return results;
        }),
        install: krl_stdlib_1.krl.Action(["url", "config"], async function install(url, config) {
            if (typeof url !== "string") {
                throw new TypeError("Expected string for url but got " + krl_stdlib_1.krl.typeOf(url));
            }
            url = normalizeUrl(url);
            const rs = await core.rsRegistry.load(url);
            await this.rsCtx.install(rs.ruleset, {
                url: url,
                config: config || {},
            });
            return rs.rid;
        }),
        uninstall: krl_stdlib_1.krl.Action(["rid"], async function uninstall(rid) {
            const pico = this.rsCtx.pico();
            if (typeof rid !== "string") {
                throw new TypeError("Expected string for rid but got " + krl_stdlib_1.krl.typeOf(rid));
            }
            const corePico = core.getPico(pico.id);
            if (!corePico) {
                throw new TypeError("Pico not found in core: " + pico.id);
            }
            const usedBy = corePico.whoUses(rid);
            if (usedBy.length > 0) {
                throw new Error(`Cannot uninstall ${rid} because ${usedBy.join(" and ")} depends on it`);
            }
            corePico.unUse(rid);
            await this.rsCtx.uninstall(rid);
        }),
        flush: krl_stdlib_1.krl.Action(["url"], async function flush(url) {
            if (typeof url !== "string") {
                throw new TypeError("Expected string for url but got " + krl_stdlib_1.krl.typeOf(url));
            }
            url = normalizeUrl(url);
            await core.rsRegistry.flush(url);
        }),
        raiseEvent: krl_stdlib_1.krl.Postlude(["domain", "name", "attrs", "forRid"], function raiseEvent(domain, name, attrs, forRid) {
            return this.rsCtx.raiseEvent(domain, name, attrs, forRid);
        }),
        event: krl_stdlib_1.krl.Action(["eci", "domain", "name", "attrs", "host"], async function event(eci, domain, name, attrs = {}, host) {
            if (host) {
                const url = `${host}/c/${eci}/event/${domain}/${name}`;
                (0, cross_fetch_1.default)(url, {
                    method: "POST",
                    headers: { "content-type": "application/json" },
                    body: krl_stdlib_1.krl.encode(attrs),
                }).catch((err) => {
                    this.log.error(err + ""); // TODO better handling
                });
                return;
            }
            // fire-n-forget event not eventWait
            const eid = await this.rsCtx.event({
                eci,
                domain,
                name,
                data: { attrs },
                time: 0,
            });
            return eid;
        }),
        eventQuery: krl_stdlib_1.krl.Action(["eci", "domain", "name", "attrs", "rid", "queryName", "args"], async function event(eci, domain, name, attrs = {}, rid, queryName, args = {}) {
            return this.rsCtx.eventQuery({
                eci,
                domain,
                name,
                data: { attrs },
                time: 0,
            }, {
                eci,
                rid,
                name: queryName,
                args,
            });
        }),
        query: krl_stdlib_1.krl.Function(["eci", "rid", "name", "args"], async function query(eci, rid, name, args = {}) {
            return this.rsCtx.query({
                eci,
                rid,
                name,
                args,
            });
        }),
        logs: krl_stdlib_1.krl.Function([], async function logs() {
            const entries = await this.getPicoLogs();
            return entries;
        }),
    };
    return module;
}
exports.default = initCtxModule;
//# sourceMappingURL=ctx.js.map