"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const cross_fetch_1 = require("cross-fetch");
const krl_stdlib_1 = require("krl-stdlib");
const qs = require("qs");
function ensureMap(arg, defaultTo) {
    return krl_stdlib_1.krl.isMap(arg) ? arg : defaultTo;
}
function stringifyQueryString(data) {
    return qs.stringify(data, {
        arrayFormat: "bracket",
        format: "RFC3986",
    });
}
async function httpBase(method, url, qs, headers, body, auth, json, form, parseJSON, dontFollowRedirect) {
    let opts = {
        method: method,
        headers: Object.assign({ "user-agent": "pico-engine-core-http" }, ensureMap(headers, {})),
    };
    if (auth && krl_stdlib_1.krl.isString(auth.username)) {
        const authHeader = `${auth.username}:${krl_stdlib_1.krl.isString(auth.password) ? auth.password : ""}`;
        opts.headers["Authorization"] =
            "Basic " + Buffer.from(authHeader).toString("base64");
    }
    if (qs) {
        url = url + "?" + stringifyQueryString(qs);
    }
    if (body) {
        opts.body = krl_stdlib_1.krl.toString(body);
    }
    else if (json) {
        opts.body = krl_stdlib_1.krl.encode(json);
        if (!opts.headers["content-type"]) {
            opts.headers["content-type"] = "application/json";
        }
    }
    else if (form) {
        opts.body = stringifyQueryString(form);
        if (!opts.headers["content-type"]) {
            opts.headers["content-type"] = "application/x-www-form-urlencoded";
        }
    }
    if (dontFollowRedirect) {
        opts.redirect = "manual";
    }
    return (0, cross_fetch_1.default)(url, opts).then(async (res) => {
        const headers = {};
        res.headers.forEach((v, k) => {
            headers[k] = v;
        });
        let r = {
            content: await res.text(),
            content_type: res.headers.get("content-type"),
            content_length: parseInt(res.headers.get("content-length") || "", 10) || 0,
            headers,
            status_code: res.status,
            status_line: res.statusText,
        };
        if (parseJSON === true) {
            try {
                r.content = krl_stdlib_1.krl.decode(r.content);
            }
            catch (e) {
                // just leave the content as is
            }
        }
        return r;
    });
}
function mkMethod(core, method, canAlsoBeUsedAsAFunction = false) {
    const params = [
        // NOTE: order is significant so it's a breaking API change to change argument ordering
        "url",
        "qs",
        "headers",
        "body",
        "auth",
        "json",
        "form",
        "parseJSON",
        "autoraise",
        "autosend",
        "dontFollowRedirect",
    ];
    const fn = async function (url, qs, headers, body, auth, json, form, parseJSON, autoraise, autosend, dontFollowRedirect) {
        if (!url) {
            throw new Error("http:" + method.toLowerCase() + " needs a url string");
        }
        if (!krl_stdlib_1.krl.isString(url)) {
            throw new TypeError("http:" +
                method.toLowerCase() +
                " was given " +
                krl_stdlib_1.krl.toString(url) +
                " instead of a url string");
        }
        if (autosend) {
            autosend = core.picoFramework.cleanEvent(autosend);
        }
        const httpPromise = httpBase(method, url, qs, headers, body, auth, json, form, parseJSON, dontFollowRedirect);
        if (autosend) {
            httpPromise
                .then((resp) => {
                return core.picoFramework.event(Object.assign(Object.assign({}, autosend), { data: {
                        attrs: Object.assign(Object.assign({}, autosend.data.attrs), resp),
                    } }));
            })
                .catch((error) => {
                this.log.error("http autosend error", { error });
            });
            return;
        }
        const r = await httpPromise;
        if (autoraise) {
            this.rsCtx.raiseEvent("http", method.toLowerCase(), Object.assign(Object.assign({}, r), { label: krl_stdlib_1.krl.toString(autoraise) }));
        }
        return r;
    };
    return canAlsoBeUsedAsAFunction
        ? krl_stdlib_1.krl.ActionFunction(params, fn)
        : krl_stdlib_1.krl.Action(params, fn);
}
function initHttpModule(core) {
    const module = {
        get: mkMethod(core, "GET", true),
        head: mkMethod(core, "HEAD", true),
        post: mkMethod(core, "POST"),
        put: mkMethod(core, "PUT"),
        patch: mkMethod(core, "PATCH"),
        delete: mkMethod(core, "DELETE"),
    };
    return module;
}
exports.default = initHttpModule;
//# sourceMappingURL=http.js.map