"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.initScheduleModule = exports.Scheduler = void 0;
const cuid = require("cuid");
const krl_stdlib_1 = require("krl-stdlib");
const _ = require("lodash");
const longTimeout = require("long-timeout"); // makes it possible to have a timeout longer than 24.8 days (2^31-1 milliseconds)
const nodeSchedule = require("node-schedule");
class Scheduler {
    constructor(scheduleJob, setTimeout, clearTimeout, now) {
        this.futures = [];
        this.crons = {};
        this.scheduleJob = scheduleJob || nodeSchedule.scheduleJob;
        this.setTimeout = setTimeout || longTimeout.setTimeout;
        this.clearTimeout = clearTimeout || longTimeout.clearTimeout;
        this.now = now || Date.now;
    }
    addFuture(id, time, handler) {
        this.futures.push({ id, time, handler });
        this.futures.sort((a, b) => a.time - b.time);
        this.update();
    }
    addCron(id, timespec, handler) {
        this.crons[id] = this.scheduleJob(timespec, handler);
    }
    remove(id) {
        if (this.crons[id]) {
            this.crons[id].cancel();
            delete this.crons[id];
        }
        else {
            let found = false;
            this.futures = this.futures.filter((future) => {
                if (future.id === id) {
                    found = true;
                    return false;
                }
                return true;
            });
            if (found) {
                this.update();
            }
        }
    }
    update() {
        this.clearTimeout(this.currTimeout);
        const next = this.futures[0];
        if (!next) {
            return;
        }
        // Execute the event by milliseconds from now.
        // If it's in the past it will happen on the next tick
        this.currTimeout = this.setTimeout(() => this.onTime(), next.time - this.now());
    }
    onTime() {
        const now = this.now();
        this.futures = this.futures.filter((future) => {
            if (future.time <= now) {
                future.handler();
                return false; // remove it now that it's done
            }
            return true;
        });
    }
}
exports.Scheduler = Scheduler;
function initScheduleModule(pf) {
    const scheduler = new Scheduler();
    function addScheduledEvent(rid, sEvent) {
        if (sEvent.type === "at") {
            scheduler.addFuture(sEvent.id, sEvent.time, async () => {
                let pico;
                try {
                    pico = pf.getPico(sEvent.event.eci);
                }
                catch (err) { }
                if (pico) {
                    // TODO wrap in pico transaction
                    const schedule = (await pico.getEnt(rid, "_schedule")) || {};
                    const job = schedule[sEvent.id];
                    if (job) {
                        await pico.putEnt(rid, "_schedule", _.omit(schedule, sEvent.id));
                        pico.event(sEvent.event);
                    }
                }
            });
        }
        else if (sEvent.type === "repeat") {
            scheduler.addCron(sEvent.id, sEvent.timespec, async () => {
                let found = false;
                let pico;
                try {
                    pico = pf.getPico(sEvent.event.eci);
                }
                catch (err) { }
                if (pico) {
                    // TODO wrap in pico transaction
                    const schedule = (await pico.getEnt(rid, "_schedule")) || {};
                    const job = schedule[sEvent.id];
                    if (job) {
                        found = true;
                        pico.event(sEvent.event);
                    }
                }
                if (!found) {
                    scheduler.remove(sEvent.id);
                }
            });
        }
    }
    async function addToSchedule(ctx, sEvent) {
        const schedule = (await ctx.rsCtx.getEnt("_schedule")) || {};
        schedule[sEvent.id] = sEvent;
        await ctx.rsCtx.putEnt("_schedule", schedule);
        addScheduledEvent(ctx.rsCtx.ruleset.rid, sEvent);
        return sEvent;
    }
    function convertToPosixTimestamp(time) {
        if (krl_stdlib_1.krl.isNumber(time)) {
            return time;
        }
        if (typeof time === "string") {
            try {
                const d = new Date(time);
                time = d.getTime();
                if (krl_stdlib_1.krl.isNumber(time)) {
                    return time;
                }
            }
            catch (err) { }
        }
        return null;
    }
    const module = {
        at: krl_stdlib_1.krl.Postlude(["time", "eci", "domain", "name", "attrs"], function (time, eci, domain, name, attrs) {
            const posixTime = convertToPosixTimestamp(time);
            if (posixTime === null || !krl_stdlib_1.krl.isNumber(posixTime)) {
                throw new TypeError("schedule .. at expected a timestamp but got " + krl_stdlib_1.krl.typeOf(time));
            }
            return addToSchedule(this, {
                id: cuid(),
                type: "at",
                time: posixTime,
                event: {
                    eci,
                    domain,
                    name,
                    data: { attrs },
                    time: 0,
                },
            });
        }),
        repeat: krl_stdlib_1.krl.Postlude(["timespec", "eci", "domain", "name", "attrs"], function (timespec, eci, domain, name, attrs) {
            return addToSchedule(this, {
                id: cuid(),
                type: "repeat",
                timespec,
                event: {
                    eci,
                    domain,
                    name,
                    data: { attrs },
                    time: 0,
                },
            });
        }),
        list: krl_stdlib_1.krl.Function([], async function () {
            const schedule = (await this.rsCtx.getEnt("_schedule")) || {};
            return Object.values(schedule);
        }),
        get: krl_stdlib_1.krl.Function(["id"], async function (id) {
            const schedule = (await this.rsCtx.getEnt("_schedule")) || {};
            return schedule[id] || null;
        }),
        remove: krl_stdlib_1.krl.Action(["id"], async function (id) {
            await this.rsCtx.putEnt("_schedule", _.omit(this.rsCtx.getEnt("_schedule") || {}, id));
            scheduler.remove(id);
        }),
        clear: krl_stdlib_1.krl.Action([], async function () {
            const schedule = (await this.rsCtx.getEnt("_schedule")) || {};
            const ids = Object.keys(schedule);
            await this.rsCtx.delEnt("_schedule");
            for (const id of ids) {
                scheduler.remove(id);
            }
        }),
    };
    return {
        module,
        start() {
            return new Promise((resolve, reject) => {
                const s = pf.db.createReadStream({
                    gte: ["entvar"],
                    lte: ["entvar", undefined], // charwise sorts with null at the bottom and undefined at the top
                });
                s.on("error", reject);
                s.on("end", () => resolve(undefined));
                s.on("data", (data) => {
                    if (data.key[3] === "_schedule") {
                        const rid = data.key[2];
                        const value = data.value;
                        for (const sEvent of Object.values(value)) {
                            addScheduledEvent(rid, sEvent);
                        }
                    }
                });
            });
        },
    };
}
exports.initScheduleModule = initScheduleModule;
//# sourceMappingURL=schedule.js.map